/*

  yyycpu - a virtual machine
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "../yyycpudisasm/disassemble.hpp"
#include "../common/hexconvert.hpp"
#include "../common/membunch.hpp"
#include "../common/fatal.hpp"

#include "machcore.h"

#include <gtk/gtk.h>
#include <signal.h>
#include <stdio.h>
#include <termios.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

bool running;
membunch *bunch=0;

void debug(); // a forward...

void machcore_illegalinstruction()
{
  printf("illegal ins\n");
  debug();
  running=false;
};

void machcore_illegalmemoryaccess()
{
  printf("illegal memaccess\n");
  debug();
  running=false;
};

#define IMAGE_WIDTH      160
#define IMAGE_HEIGHT     160

guchar rgbbuf[IMAGE_WIDTH * IMAGE_HEIGHT * 3];
guchar rgbbufback[IMAGE_WIDTH * IMAGE_HEIGHT * 3];
GtkWidget *darea;

gboolean on_darea_expose (GtkWidget *widget,GdkEventExpose *event,gpointer user_data)
{
  gdk_draw_rgb_image (widget->window, widget->style->fg_gc[GTK_STATE_NORMAL],
		      0, 0, IMAGE_WIDTH, IMAGE_HEIGHT,
		      GDK_RGB_DITHER_MAX, rgbbuf, IMAGE_WIDTH * 3);
  return true;
}

void destroy( GtkWidget *widget, gpointer   data )
{
  debug();
  gtk_main_quit();
}





inline void setpixel(const int32 x, const int32 y, const bool iswhite)
{
  if ((x<0)||(x>=160)||(y<0)||(y>=160)) return;
  guchar val=iswhite?255:0;
  rgbbufback[(y*160+x)*3]=val;
  rgbbufback[(y*160+x)*3+1]=val;
  rgbbufback[(y*160+x)*3+2]=val;
};

void drawrectangle (const int x1, const int y1, const int x2, const int y2, const bool white)
{
  for (int y=y1;y<=y2;y++)
    for (int x=x1;x<=x2;x++)
      setpixel(x,y,white);
};

void drawsprite (uint16 posx, uint16 posy, uint16 width, uint16 height, uint32 address)
{
  for (int y=0;y<height;++y)
    for (int x=0;x<width;++x)
      setpixel(posx+x,posy+y,(machcore_memory[address+y*width+x]==' ') ? true : false);
  printf("%d %d %d %d %d\n",posx,posy,width,height,address);
};

void updatescr()
{
  memcpy(rgbbuf,rgbbufback,160*160*3);
  gtk_widget_draw(darea,0);
};
  
u_int32_t machcore_in(const u_int32_t port)
{
  int c;
  switch (port) {
  case 0x50:
    c=getchar();
    if (c==EOF) return 0xFFFFFFFF;
    return c;
  default:
    printf("IO_in(%s) unknown port.\n",uint32toglobalhex32(port));
    return 0;
  };
};

void machcore_out(const u_int32_t port, const u_int32_t value)
{
  static int x1=0,y1=0,x2=0,y2=0;
  static bool white=true;
  static uint16 posx=0, posy=0;
  static uint16 width=0, height=0;
  static uint32 spriteaddress=0;
  switch (port) {
  case 0x10: running=false; printf("SYSTEM SHUTDOWN INITIATED\n"); break;
  case 0x50:
    printf("%c",value);fflush(stdout);
    break;
  case 0x20: x1 = value; break;
  case 0x21: y1 = value; break;
  case 0x22: x2 = value; break;
  case 0x23: y2 = value; break;
  case 0x24: setpixel (x1, y1, white); break;
  case 0x25: setpixel (x2, y2, white); break;
  case 0x26: drawrectangle (x1,y1,x2,y2,white); break;
  case 0x27: white=value?true:false; break;
  case 0x28: updatescr(); break;
  case 0x1000: posx=value; break;
  case 0x1001: posy=value; break;
  case 0x1002: width=value; break;
  case 0x1003: height=value; break;
  case 0x1004: spriteaddress=value; break;
  case 0x1005: drawsprite (posx,posy,width,height,spriteaddress); break;
  default:
    printf("IO_out(%s",uint32toglobalhex32(port));
    printf(",%s) unknown port.\n",uint32toglobalhex32(value));
    break;
  };
};

gint myidle (gpointer)
{
  disas_addr=machcore_PC;
  //  printf("%s\n",disassemble().c_str());
  machcore_execute(1000);
  //  debug();
  //char c;read(0,&c,1);
  if (!running) {
    debug();
    return false;
    gtk_main_quit();
  };
  return TRUE; 
};

void debug()
{
  char *reg[] = {
    "R0 ",
    "R1 ",
    "R2 ",
    "R3 ",
    "R4 ",
    "R5 ",
    "R6 ",
    "R7 ",
    "R8 ",
    "R9 ",
    "R10",
    "R11",
    "R12",
    "R13",
    "R14",
    "R15"
  };
  for (uint8 y=0;y<4;y++) {
    for (uint8 x=0;x<4;x++)
      printf("%s = %s   ", reg[y*4+x], uint32toglobalhex32(machcore_regs[y*4+x]));
    printf("\n");
  };
  printf(machcore_interrupt_enable ? "interrupts enabled\n" : "interrupts DISABLED\n");
  printf("PC=%d\n",machcore_PC);
};



int main(int argc,char *argv[])
{
  machcore_initialize(30000);
  bunch = new membunch(machcore_memory,machcore_memsize);
  disas_membunch=bunch;
  running=true;

  if (argc!=2) fatal("usage: machine binfile");
  FILE *f = fopen(argv[1],"r");
  if (!f) fatal ("can't open file");
  int val; uint32 i=0;
  while ((val=fgetc(f))!=EOF) {
    machcore_memory[i++]=val;
    if (i==machcore_memsize) fatal("cannot load program. memory too small");
  };
  if (fclose(f)) fatal("can't close file");

  GtkWidget *window;
  gtk_init (&argc, &argv);
  gdk_rgb_init();
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  darea = gtk_drawing_area_new();
  gtk_widget_set_usize(window,160,160);
  gtk_drawing_area_size (GTK_DRAWING_AREA (darea), IMAGE_WIDTH, IMAGE_HEIGHT);
  gtk_container_add (GTK_CONTAINER (window), darea);
  gtk_signal_connect (GTK_OBJECT (darea), "expose-event", GTK_SIGNAL_FUNC (on_darea_expose), NULL);
  gtk_signal_connect (GTK_OBJECT (window), "destroy", GTK_SIGNAL_FUNC (destroy), NULL);
  gtk_idle_add(myidle,0);
  gtk_widget_show_all (window);
  gtk_window_set_policy(GTK_WINDOW(window),false,false,false);
  memset(rgbbuf,0,160*160*3);
  memset(rgbbufback,0,160*160*3);

  gtk_main();

  delete bunch;
  machcore_deinitialize();
  return 0;
};
